import logging
import contextlib
from functools import cached_property
from pathlib import Path
from typing import Generator, List, Optional
from elx.singer import Singer, require_install
from elx.catalog import Stream
from elx.json_temp_file import json_temp_file
from subprocess import Popen, PIPE


class Tap(Singer):
    def discover(self, config_path: Path) -> dict:
        """
        Run the tap in discovery mode.

        Args:
            config_path (Path): Path to the config file.

        Returns:
            dict: The catalog.
        """
        logging.debug(f"Discovering {self.executable} with {config_path}")
        return self.run(["--config", str(config_path), "--discover"])

    @cached_property
    def catalog(self) -> dict:
        """
        Discover the catalog.

        Returns:
            Catalog: The catalog as a Pydantic model.
        """
        with json_temp_file(self.config) as config_path:
            catalog = self.discover(config_path)
            return catalog

    def filtered_catalog(
        self, catalog: dict, streams: Optional[List[str]] = None
    ) -> dict:
        """
        Filter the catalog.

        Args:
            catalog (dict): The catalog.
            streams (Optional[List[str]], optional): The streams to filter on. Defaults to None.

        Returns:
            dict: The filtered catalog.
        """
        if not streams:
            return catalog

        return {
            "streams": [
                {
                    **stream,
                    "selected": stream["tap_stream_id"] in streams,
                    "metadata": stream["metadata"]
                    + [
                        {
                            "metadata": {
                                "selected": stream["tap_stream_id"] in streams,
                            },
                            "breadcrumb": [],
                        }
                    ],
                }
                for stream in catalog["streams"]
            ]
        }

    @property
    def streams(self) -> list[Stream]:
        """
        Get the streams from the catalog.

        Returns:
            list[Stream]: The streams.
        """
        return [Stream(**stream) for stream in self.catalog["streams"]]

    @contextlib.contextmanager
    @require_install
    def process(
        self,
        state: dict = {},
        streams: Optional[List[str]] = None,
    ) -> Generator[Popen, None, None]:
        """
        Run the tap process.

        Returns:
            Popen: The tap process.
        """
        catalog = self.filtered_catalog(catalog=self.catalog, streams=streams)

        with json_temp_file(self.config) as config_path:
            with json_temp_file(catalog) as catalog_path:
                with json_temp_file(state) as state_path:
                    yield Popen(
                        [
                            self.executable,
                            "--config",
                            str(config_path),
                            "--catalog",
                            str(catalog_path),
                            "--state",
                            str(state_path),
                        ],
                        stdout=PIPE,
                        stderr=PIPE,
                    )



from elx import Runner
from elx.catalog import Stream

def dagster_safe_name(name: str) -> str:
    return name.replace("-", "_")

def schema(stream: Stream) -> str:
    # return stream.stream_schema.properties
    return '\n'.join(stream.stream_schema.properties.keys())

def generate_description(runner: Runner, stream: Stream) -> str:
    return f"""
        Load stream '{stream.name}' from '{runner.tap.executable}' into '{runner.target.executable}'.

        Refresh method: {stream.replication_method}
    """


def require_install(func):
    """
    Decorator to check if the executable is installed.
    """

    def wrapper(self, *args, **kwargs):
        if not self.is_installed:
            self.install()

        return func(self, *args, **kwargs)

    return wrapper


def interpolate_in_config(config: dict, interpolation: dict) -> dict:
    """
    Interpolate a value in the config. Recurse through the config and use format strings to interpolate the value.

    Args:
        config (dict): The config.
        interpolation (dict): The config to interpolate with, E.g. {"key": "value"}.

    Returns:
        dict: The interpolated config.
    """

    def _interpolate(value: str) -> str:
        if isinstance(value, str):
            return value.format(**interpolation)
        elif isinstance(value, list):
            return [_interpolate(item) for item in value]
        elif isinstance(value, dict):
            return {key: _interpolate(value) for key, value in value.items()}
        else:
            return value

    return {key: _interpolate(value) for key, value in config.items()}



from subprocess import PIPE, Popen
from typing import Generator, Optional
from elx.singer import Singer, require_install
from elx.json_temp_file import json_temp_file
import contextlib


class Target(Singer):
    @contextlib.contextmanager
    @require_install
    def process(
        self,
        tap_process: Popen,
        config_interpolation: Optional[dict] = {},
    ) -> Generator[Popen, None, None]:
        """
        Run the tap process.

        Args:
            tap_process (Popen): The process where the tap is running.
                Is used to pipe the output to the target.
            config_interpolation (Optional[dict], optional): Values that can be
                used in the config of the tap or target. Defaults to {}.

        Returns:
            Popen: The tap process.
        """
        with json_temp_file(self.config) as config_path:
            yield Popen(
                [
                    self.executable,
                    "--config",
                    str(config_path),
                ],
                stdin=tap_process.stdout,
                stdout=PIPE,
                stderr=PIPE,
            )



import json
import logging
import subprocess
import hashlib
from distutils.spawn import find_executable
from functools import cached_property
from typing import Generator, Optional
from pipx.commands.install import install
from pipx.commands.uninstall import uninstall
from pipx.commands.common import package_name_from_spec
from pipx.constants import LOCAL_BIN_DIR
from elx.exceptions import DecodeException, PipxInstallException
from elx.utils import require_install, interpolate_in_config

PYTHON = "python3"


class Singer:
    def __init__(
        self,
        spec: str,
        executable: Optional[str] = None,
        config: dict = {},
    ):
        self.spec = spec
        self._executable = executable
        self._config = config

    @property
    def config(self) -> dict:
        """
        Get the config for this plugin.
        """
        config = self._config

        # If the config is a callable, call it and return the result.
        if callable(self._config):
            config = self._config()

        # If there is a runner attribute, interpolate the config.
        if hasattr(self, "runner"):
            config = interpolate_in_config(
                config=config,
                interpolation=self.runner.interpolation_values,
            )

        return config

    @cached_property
    def executable(self) -> str:
        """
        Get the package name for this plugin.
        """
        if self._executable:
            return self._executable

        return package_name_from_spec(
            package_spec=self.spec,
            python=PYTHON,
            pip_args=[],
            verbose=False,
        )

    @cached_property
    def hash_key(self) -> str:
        """
        Generate a md5 hash key for this plugin, based on the executable, spec and config.
        """
        return hashlib.md5(
            json.dumps(
                {
                    "executable": self.executable,
                    "spec": self.spec,
                    "config": self.config,
                }
            ).encode()
        ).hexdigest()

    @property
    def is_installed(self) -> bool:
        """
        Check if the executable is installed.

        Returns:
            bool: True if the executable is installed, False otherwise.
        """
        return find_executable(self.executable) is not None

    def install(self) -> None:
        """
        Install the executable using pipx.
        """
        logging.info(f"Installing {self.executable}...")

        try:
            subprocess.run(
                [
                    "pipx",
                    "install",
                    self.spec or self.executable,
                    "--force",
                ],
                capture_output=True,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            raise PipxInstallException(e.stderr.decode())

    @require_install
    def run(self, args: list) -> dict:
        """
        Run the executable with the given arguments.

        Args:
            args (list): The arguments to pass to the executable.

        Returns:
            dict: The JSON output of the executable.

        Raises:
            DecodeException: If the JSON output of the executable is not valid.
        """
        result = subprocess.Popen(
            [
                self.executable,
                *args,
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        stdout, stderr = result.communicate()

        # If any of the processes exited with a non-zero exit code,
        # raise an exception.
        if result.returncode != 0:
            raise DecodeException(f"Error running {self.executable}: {stderr.decode()}")

        # Else try to parse the JSON output.
        try:
            return json.loads(stdout)
        except json.decoder.JSONDecodeError as e:
            raise DecodeException(
                f"Error parsing json: {e.msg} at position {e.pos} in {e.doc} \n\n {stderr.decode()}"
            )



from typing import List, Optional
from pydantic import BaseModel, Field


class Schema(BaseModel):
    properties: dict


class Stream(BaseModel):
    tap_stream_id: str
    replication_method: Optional[str] = "FULL_REFRESH"
    key_properties: List[str]
    stream_schema: Schema = Field(alias="schema")

    @property
    def name(self) -> str:
        return self.tap_stream_id

    @property
    def safe_name(self) -> str:
        return self.name.replace("-", "_")



import contextlib
import json
from pathlib import Path
import tempfile
from typing import Generator


@contextlib.contextmanager
def json_temp_file(content: dict) -> Generator[Path, None, None]:
    """
    Write a json object to a temporary file. The file is deleted after the
    context manager exits.

    Args:
        content (dict): The object to write to the file.

    Yields:
        Path: Path to the catalog.
    """
    try:
        # Use tempfile to create a temporary config file.
        with tempfile.NamedTemporaryFile(
            mode="w",
            suffix=".json",
            delete=False,
        ) as catalog_file:
            # Write the config attribute to the config file.
            catalog_file.write(json.dumps(content))
        # Yield the path to the config file.
        yield Path(catalog_file.name)
    finally:
        # Always delete the catalog file.
        Path(catalog_file.name).unlink()



from abc import ABC, abstractproperty
from functools import cache
from pathlib import Path
import json
from sre_parse import State
from typing import Any, Dict
from smart_open import open
import os
from functools import cached_property


class StateClient(ABC):
    def __init__(self, base_path: str):
        self.base_path = base_path

    @property
    def client(self) -> Any:
        """
        Returns:
            Any: The client to use for interacting with the state store.
        """
        raise NotImplementedError

    @property
    def params(self) -> Dict[str, Any]:
        """
        Returns:
            Dict[str, Any]: The parameters to pass to smart_open.open.
        """
        return {
            "client": self.client,
        }


class S3StateClient(StateClient):
    """
    A state client for S3 state stores.
    """

    @cached_property
    def client(self):
        import boto3

        session = boto3.Session(
            aws_access_key_id=os.environ["AWS_ACCESS_KEY_ID"],
            aws_secret_access_key=os.environ["AWS_SECRET_ACCESS_KEY"],
        )
        return session.client("s3")


class AzureStateClient(StateClient):
    """
    A state client for Azure Blob Storage state stores.
    """

    @cached_property
    def client(self):
        from azure.storage.blob import BlobServiceClient

        return BlobServiceClient.from_connection_string(
            os.environ["AZURE_STORAGE_CONNECTION_STRING"]
        )


class GCSStateClient(StateClient):
    """
    A state client for Google Cloud Storage state stores.
    """

    @cached_property
    def client(self):
        from google.cloud.storage import Client
        from google.auth.credentials import Credentials

        if "GOOGLE_APPLICATION_CREDENTIALS" in os.environ:
            service_account_path = os.environ["GOOGLE_APPLICATION_CREDENTIALS"]
            return Client.from_service_account_json(service_account_path)
        elif "GOOGLE_API_TOKEN" in os.environ:
            token = os.environ["GOOGLE_API_TOKEN"]
            credentials = Credentials(token=token)
            return Client(credentials=credentials)
        else:
            raise Exception("No credentials found for Google Cloud Storage")


class LocalStateClient(StateClient):
    """
    A state client for local (and all other) state stores.
    """

    @property
    def params(self) -> dict:
        return {}


def state_client_factory(base_path: str) -> StateClient:
    if base_path.startswith("s3://"):
        return S3StateClient(base_path)
    elif base_path.startswith("azure://"):
        return AzureStateClient(base_path)
    elif base_path.startswith("gs://"):
        return GCSStateClient(base_path)
    else:
        return LocalStateClient(base_path)


class StateManager:
    def __init__(self, base_path: str = ".") -> None:
        """
        Args:
            base_path (str): The base path to store state files in. Defaults to "./state".
        """
        self.base_path = base_path
        self.state_client = state_client_factory(base_path)

    def load(self, state_file_name: str) -> dict:
        """
        Load a state file.

        Args:
            state_file_name (str): The name of the state file to load.

        Returns:
            dict: The contents of the state file.
        """
        if not Path(f"{self.base_path}/{state_file_name}").exists():
            return {}

        with open(
            f"{self.base_path}/{state_file_name}",
            "r",
            transport_params=self.state_client.params,
        ) as state_file:
            return json.loads(state_file.read())

    def save(self, state_file_name: str, state: dict = {}) -> None:
        """
        Save a state file.

        Args:
            state_file_name (str): The name of the state file to save.
        """
        # We first merge with any existing state to ensure that we don't overwrite
        existing_state = self.load(state_file_name)
        merged_state = {**existing_state, **state}

        # Then we write the merged state to the state file
        with open(
            f"{self.base_path}/{state_file_name}",
            "wb",
            transport_params=self.state_client.params,
        ) as state_file:
            state_file.write(json.dumps(merged_state).encode("utf-8"))



import datetime
import json
import logging
import subprocess
from typing import List, Optional

from functools import cached_property
from elx.tap import Tap
from elx.target import Target
from elx import StateManager
from dotenv import load_dotenv

logging.basicConfig(level=logging.INFO)


class Runner:
    def __init__(
        self,
        tap: Tap,
        target: Target,
        state_manager: StateManager = StateManager(),
    ):
        load_dotenv()
        self.tap = tap
        self.target = target
        self.state_manager = state_manager

        # Give the tap and target access to the runner.
        self.tap.runner = self
        self.target.runner = self

    @property
    def state_file_name(self) -> str:
        return f"{self.tap.executable}-{self.target.executable}.json"

    def load_state(self) -> dict:
        return self.state_manager.load(self.state_file_name)

    def save_state(self, state: dict) -> None:
        self.state_manager.save(self.state_file_name, state)

    @cached_property
    def interpolation_values(self) -> dict:
        """
        Values that can be used in the config of the tap or target.
        """
        NOW = datetime.datetime.now()

        return {
            "NOW": NOW.isoformat(),
            "YESTERDAY": (NOW - datetime.timedelta(days=1)).isoformat(),
            "LAST_WEEK": (NOW - datetime.timedelta(days=7)).isoformat(),
            "TAP_EXECUTABLE": self.tap.executable,
            "TAP_NAME": self.tap.executable.replace("-", "_"),
            "TARGET_EXECUTABLE": self.target.executable,
            "TARGET_NAME": self.target.executable.replace("-", "_"),
        }

    def run(
        self,
        streams: Optional[List[str]] = None,
        logger: logging.Logger = None,
    ) -> None:
        state = self.load_state()

        with self.tap.process(state=state, streams=streams) as tap_process:
            with self.target.process(tap_process=tap_process) as target_process:

                def log_lines():
                    yield from iter(tap_process.stderr.readline, b"")
                    yield from iter(target_process.stderr.readline, b"")

                for line in log_lines():
                    print(line.decode("utf-8"))
                    if logger:
                        logger.info(line.decode("utf-8"))

                tap_process.stdout.close()
                stdout, stderr = target_process.communicate()

                # If any of the processes exited with a non-zero exit code,
                # raise an exception.
                if tap_process.returncode and tap_process.returncode != 0:
                    raise subprocess.CalledProcessError(
                        tap_process.returncode, tap_process.args
                    )
                if target_process.returncode and target_process.returncode != 0:
                    raise subprocess.CalledProcessError(
                        target_process.returncode, target_process.args
                    )

                state = json.loads(stdout.decode("utf-8"))
                self.save_state(state)


if __name__ == "__main__":
    tap = Tap(
        executable="tap-smoke-test",
        spec="git+https://github.com/meltano/tap-smoke-test.git",
        config={
            "streams": [
                {
                    "stream_name": "stream-one",
                    "input_filename": "https://gitlab.com/meltano/tap-smoke-test/-/raw/main/demo-data/animals-data.jsonl",
                },
                {
                    "stream_name": "stream-two",
                    "input_filename": "https://gitlab.com/meltano/tap-smoke-test/-/raw/main/demo-data/animals-data.jsonl",
                },
            ],
        },
    )
    target = Target(
        spec="git+https://github.com/estrategiahq/target-parquet.git",
        executable="target-parquet",
        config={
            "destination_path": "/tmp",
        },
    )
    runner = Runner(tap, target, StateManager("/tmp"))
    runner.run("stream-one")



import logging
import warnings

warnings.filterwarnings("ignore", category=DeprecationWarning)

from elx.state import StateManager
from elx.tap import Tap
from elx.target import Target
from elx.runner import Runner

logger = logging.getLogger("pipx")
logger.setLevel(logging.CRITICAL)



from typing import Generator, List
from elx import Runner
from dagster import (
    AssetsDefinition,
    Nothing,
    OpExecutionContext,
    Output,
    multi_asset,
    AssetOut,
    get_dagster_logger,
)
from elx.extensions.dagster.utils import dagster_safe_name, generate_description

logger = get_dagster_logger()


def load_assets(runner: Runner) -> List[AssetsDefinition]:
    """
    Load the assets for a runner, each asset represents one tap target combination.

    Args:
        runner (Runner): The runner to extract from.

    Returns:
        List[AssetsDefinition]: The assets.
    """

    def run_factory(runner: Runner) -> callable:
        """
        Create a run function for a runner.

        Args:
            runner (Runner): The runner to create a run function for.

        Returns:
            callable: The run function that gets executed by Dagster.
        """

        def run(context: OpExecutionContext) -> Generator[Output, None, None]:
            """
            Run a tap target combination.

            Args:
                context (OpExecutionContext): The context to run in.

            Yields:
                Generator[Output, None, None]: The names of the selected outputs.
            """
            # Execute the runner and yield the selected outputs.
            runner.run(
                streams=list(context.selected_output_names),
                logger=logger,
            )

            for context_output_name in context.selected_output_names:
                yield Output(
                    value=Nothing,
                    output_name=context_output_name,
                    metadata={
                        "state_path": f"{runner.state_manager.base_path}/{runner.state_file_name}",
                        "state": runner.load_state(),
                    },
                )

        return run

    return [
        multi_asset(
            name=f"run_{dagster_safe_name(runner.tap.executable)}_{dagster_safe_name(runner.target.executable)}",
            outs={
                dagster_safe_name(stream.name): AssetOut(
                    is_required=False,
                    description=generate_description(runner=runner, stream=stream),
                    key_prefix=dagster_safe_name(runner.tap.executable),
                    code_version=runner.tap.hash_key,
                )
                for stream in runner.tap.streams
            },
            can_subset=True,
            group_name=dagster_safe_name(runner.tap.executable),
            compute_kind="python",
        )(run_factory(runner))
    ]
